%% (logic) AUFLIA
%% (info) :source 
%% Binary seach algorithm
%% ======================
%% 
%% @pre: 0 ≤ l ∧ u < length(a) ∧ sorted(a, 0, length(a)-1)
%% @post: res ↔ ∃ i (l ≤ i ∧ i ≤ u ∧ a[i] = e)
%% def binSearch(a: Array[Int], l: Int, u: Int, e: Int) =
%%   if (l > u)
%%      false
%%   else {
%%      val m = (l + u) / 2
%%      if (a[m] == e)
%%         true
%%      else if (a[m] < e)
%%         binSearch(a, m+1, u, e)
%%      else 
%%         binSearch(a, l, m-1, e)
%%   }
%% 
%% Partial correctness:
%% 
%% Proof of postcondition by induction, assuming precondition:
%% ∀ l:Int, u:Int, a:Array[Int], e:Int
%%   0 ≤ l ∧ u < length(a) ∧ sorted(a, 0, length(a)-1) →
%%   let res =
%%       if l > u then
%%         false
%%       else
%%         let m = (l + u) / 2 
%%          in if a[m] = e then
%%                true
%%             else if a[m] < e
%%                let res1 =         // res1 is result of recursive call, by I.H.
%%                    let l = m+1
%%                     in ∃ i:Int (l ≤ i ∧ i ≤ u ∧ a[i] = e) // I.H.
%%                 in res1 // result res is just res1
%%             else
%%                let res1 =        // res1 is result of recursive call, by I.H.
%%                    let u = m-1
%%                     in ∃ i:Int (l ≤ i ∧ i ≤ u ∧ a[i] = e) // I.H.
%%                 in res1 // result res is just res1
%%    in res ↔ ∃ i (l ≤ i ∧ i ≤ u ∧ a[i] = e) // prove postcondition
%% 
%% Proof precondition of induction hypotheses:
%% ∀ l:Int, u:Int, a:Array[Int], e:Int
%%   0 ≤ l ∧ u < length(a) ∧ sorted(a, 0, length(a)-1) →
%%   if l > u then
%%     true
%%   else
%%     let m = (l + u) / 2 
%%      in if a[m] = e then
%%            true
%%         else if a[m] < e
%%            let l = m+1
%%             in 0 ≤ l ∧ u < length(a) ∧ sorted(a, 0, length(a)-1)
%%         else
%%            let u = m-1
%%             in 0 ≤ l ∧ u < length(a) ∧ sorted(a, 0, length(a)-1)
%% 

include('swv-array-axioms.p').

%% Declarations:
tff(length, type, length: 'Array[Int,Int]' > $int).
tff(div2, type, div2: $int > $int).
tff(sorted, type, sorted: ('Array[Int,Int]' * $int * $int) > $o).

%% Definition of divisibility by 2:
%% ∀ a:Int res:Int ((((2 * res) ≤ a) ∧ ((2 * (res + 1)) > a)) = (div2(a) = res))
tff(formula, axiom,
  ( ! [A:$int, Res:$int] :
      ( ( $lesseq($product(2, Res), A)
	& $greater($product(2, $sum(Res, 1)), A))
    <=> (div2(A) = Res)))).

%% The length of arrays is greater or equal to 0:
%% ∀ a:Array[Int, Int] (length(a) ≥ 0)
tff(formula, axiom,
  ( ! [A:'Array[Int,Int]'] :
      $greatereq(length(A), 0))).

%% Sortedness of arrays:
%% ∀ a:Array[Int, Int] l:Int u:Int (sorted(a, l, u) = ∀ i:Int j:Int (((l ≤ i) ∧ (i < j) ∧ (j ≤ u)) ⇒ (a[i] ≤ a[j])))
tff(formula, axiom,
  ( ! [A:'Array[Int,Int]', L:$int, U:$int] :
      ( sorted(A, L, U)
    <=> ( ! [I:$int, J:$int] :
	    ( ( $lesseq(L, I)
	      & $less(I, J)
	      & $lesseq(J, U))
	   => $lesseq('select:(Array[Int,Int]*Int)>Int'(A, I), 'select:(Array[Int,Int]*Int)>Int'(A, J))))))).

%% (Negated) conjecture according to first proof obligation above:
%% ¬∀ l:Int u:Int a:Array[Int, Int] e:Int (((0 ≤ l) ∧ (u < length(a)) ∧ sorted(a, 0, (length(a) - 1))) ⇒ let res=(if (l > u) false else let m=div2((l + u)) in (if (a[m] = e) true else (if (a[m] < e) let res1=let l=(m + 1) in ∃ i:Int ((l ≤ i) ∧ (i ≤ u) ∧ (a[i] = e)) in res1 else let res1=let u=(m - 1) in ∃ i:Int ((l ≤ i) ∧ (i ≤ u) ∧ (a[i] = e)) in res1))) in (res = ∃ i:Int ((l ≤ i) ∧ (i ≤ u) ∧ (a[i] = e))))
tff(formula, conjecture,
  (  ( ! [L:$int, U:$int, A:'Array[Int,Int]', E:$int] :
	 ( ( $lesseq(0, L)
	   & $less(U, length(A))
	   & sorted(A, 0, $difference(length(A), 1)))
	=> ( ( ( $greater(L, U)
	      => $false)
	     & ( (~ $greater(L, U))
	      => ( ? [M_2:$int] :
		     ( (M_2 = div2($sum(L, U)))
		     & ( ( ('select:(Array[Int,Int]*Int)>Int'(A, M_2) = E)
			=> $true)
		       & ( ('select:(Array[Int,Int]*Int)>Int'(A, M_2) != E)
			=> ( ( $less('select:(Array[Int,Int]*Int)>Int'(A, M_2), E)
			    => ( ? [L_4:$int] :
				   ( (L_4 = $sum(M_2, 1))
				   & ( ? [I:$int] :
					 ( $lesseq(L_4, I)
					 & $lesseq(I, U)
					 & ('select:(Array[Int,Int]*Int)>Int'(A, I) = E))))))
			   & ( (~ $less('select:(Array[Int,Int]*Int)>Int'(A, M_2), E))
			    => ( ? [U_6:$int] :
				   ( (U_6 = $difference(M_2, 1))
				   & ( ? [I:$int] :
					 ( $lesseq(L, I)
					 & $lesseq(I, U_6)
					 & ('select:(Array[Int,Int]*Int)>Int'(A, I) = E)))))))))))))
	 <=> ( ? [I:$int] :
		 ( $lesseq(L, I)
		 & $lesseq(I, U)
		 & ('select:(Array[Int,Int]*Int)>Int'(A, I) = E)))))))).

